# SPDX-FileCopyrightText: Coypright © 2024 Shooting Soul Ventures, LLC <jg@shootingsoul.com>
# SPDX-License-Identifier: MIT

from configsweep.affiliate import AffiliateClass
from configsweep.sweep import Sweep
from configsweep.sweep_combination import SweepCombination
from configsweep.sweeper import Sweeper

__version__ = "1.0.0"
__all__ = (__version__,
           AffiliateClass,
           Sweep,
           Sweeper,
           SweepCombination)
